<?php

namespace Drupal\plugin_form_element\Element;

use Drupal\Component\Plugin\PluginInspectionInterface;
use Drupal\Core\Form\FormState;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformState;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\Core\Plugin\PluginWithFormsInterface;
use Drupal\Core\Render\Attribute\FormElement;
use Drupal\Core\Render\Element\FormElementBase;

/**
 * Provides a plugin form element.
 *
 * This builds a plugin form for a specific plugin type. It takes complete care
 * of validating and processing the form values while respecting the plugin form
 * implementations of validate and submit methods.
 *
 * Properties:
 * - #default_value: Must not be NULL. Plugin configuration stores the plugin's
 *   id inside its configuration, and we need this to initialize the plugin
 *   itself. Therefore, it must at least contain the id of the plugin.
 *
 * Usage example:
 * @code
 * $form['my_plugin_element'] = [
 *   '#type' => 'plugin',
 *   '#title' => $this->t('Configure condition'),
 *   '#plugin_manager' => 'plugin.manager.condition',
 *   '#default_value' => [
 *     'id' => 'language',
 *     'langcodes' => ['en' => 'en'],
 *   ],
 * ];
 * @endcode
 */
#[FormElement('plugin')]
class Plugin extends FormElementBase {

  /**
   * {@inheritdoc}
   */
  public function getInfo(): array {
    $class = static::class;
    return [
      '#input' => TRUE,
      '#process' => [
        [$class, 'processPluginElement'],
        [$class, 'buildPluginForm'],
      ],
      '#element_validate' => [
        [$class, 'validatePluginForm'],
      ],
      '#theme_wrappers' => ['form_element'],
    ];
  }

  /**
   * Processes the plugin form element.
   *
   * @param array $element
   *   The form element to process.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param array $complete_form
   *   The complete form structure.
   *
   * @return array
   *   The processed element.
   */
  public static function processPluginElement(array &$element, FormStateInterface $form_state, array &$complete_form): array {
    $plugin_id = $element['#default_value']['id'] ?? NULL;

    if (!$plugin_id) {
      \Drupal::messenger()->addWarning(t('The plugin form element must at least define an id in its default configuration. See usage documentation in \Drupal\plugin_form_element\Element\Plugin.'));
      return $element;
    }

    /** @var \Drupal\Component\Plugin\PluginManagerInterface $plugin_manager */
    $plugin_manager = $element['#plugin_manager'] ? \Drupal::service($element['#plugin_manager']) : FALSE;
    if (!$plugin_manager) {
      \Drupal::messenger()->addWarning(t('The plugin form element must define a valid plugin manager. See usage documentation in \Drupal\plugin_form_element\Element\Plugin.'));
      return $element;
    }

    if (!$plugin_manager->hasDefinition($plugin_id)) {
      \Drupal::messenger()->addWarning(t('The plugin manager "@plugin_manager" has no plugin with id "@plugin_id".', [
        '@plugin_manager' => $element['#plugin_manager'],
        '@plugin_id' => $plugin_id,
      ]));
      return $element;
    }

    $configuration = $element['#default_value'] ?? [];
    $plugin = $plugin_manager->createInstance($plugin_id, $configuration);

    if (!$plugin instanceof PluginFormInterface) {
      \Drupal::messenger()->addWarning(t('The plugin instance for "@plugin_id" is not implementing \Drupal\Core\Plugin\PluginFormInterface.', [
        '@plugin_id' => $plugin_id,
      ]));
      return $element;
    }

    return array_merge([
      '#plugin_id' => $plugin_id,
      '#plugin' => $plugin,
    ], $element);
  }

  /**
   * Builds the plugin form for the processed element.
   *
   * @param array $element
   *   The form element to process.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param array $complete_form
   *   The complete form structure.
   *
   * @return array
   *   The processed element.
   */
  public static function buildPluginForm(array &$element, FormStateInterface $form_state, array &$complete_form): array {
    $sub_form_state = SubformState::createForSubform($element, $complete_form, $form_state);
    $sub_form_state->setTemporaryValue('gathered_contexts', \Drupal::service('context.repository')->getAvailableContexts());

    return static::getPluginForm($element['#plugin'])->buildConfigurationForm($element, $sub_form_state);
  }

  /**
   * Validates the plugin form.
   *
   * @param array $element
   *   The form element to process.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param array $complete_form
   *   The complete form structure.
   */
  public static function validatePluginForm(array &$element, FormStateInterface $form_state, array &$complete_form): void {
    $sub_form_state = SubformState::createForSubform($element, $complete_form, $form_state);
    static::getPluginForm($element['#plugin'])->validateConfigurationForm($element, $sub_form_state);
  }

  /**
   * {@inheritdoc}
   */
  public static function valueCallback(&$element, $input, FormStateInterface $form_state) {
    if ($input) {
      $complete_form = $form_state->getCompleteForm();
      $element = static::processPluginElement($element, $form_state, $complete_form);
      $sub_form_state = new FormState();
      $sub_form_state->setValues(array_merge($element['#plugin']->defaultConfiguration(), $input));
      static::getPluginForm($element['#plugin'])->submitConfigurationForm($element, $sub_form_state);

      // Update the values with the process configuration from the plugin.
      $input = $element['#plugin']->getConfiguration();
    }

    return $input ?: [];
  }

  /**
   * Retrieves the plugin form for a given block and operation.
   *
   * @param \Drupal\Component\Plugin\PluginInspectionInterface $plugin
   *   The plugin.
   *
   * @return \Drupal\Component\Plugin\PluginInspectionInterface
   *   The form for the plugin.
   */
  protected static function getPluginForm(PluginInspectionInterface $plugin) {
    return $plugin instanceof PluginWithFormsInterface
      ? \Drupal::service('plugin_form.factory')->createInstance($plugin, 'configure')
      : $plugin;
  }

}
