<?php

declare(strict_types=1);

namespace Drupal\plugin_form_element\Element;

use Drupal\Component\Utility\Html;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Plugin\FilteredPluginManagerInterface;
use Drupal\Core\Render\Attribute\FormElement;
use Drupal\Core\Render\Element;
use Drupal\Core\Render\Element\FormElementBase;

/**
 * Provides a plugins element.
 *
 * Properties:
 * - #plugin_manager: The service id of the plugin manager which should be used
 *   to load the config form from.
 *
 * Usage example:
 * @code
 * $form['my_plugins_element'] = [
 *   '#type' => 'plugins',
 *   '#title' => $this->t('Configure conditions'),
 *   '#plugin_manager' => 'plugin.manager.condition',
 *   '#cardinality' => -1,
 * ];
 * @endcode
 */
#[FormElement('plugins')]
class Plugins extends FormElementBase {

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = static::class;
    return [
      '#input' => TRUE,
      '#process' => [
        [$class, 'processPluginsElement'],
        [$class, 'buildPluginsForm'],
      ],
      '#theme_wrappers' => ['form_element'],
    ];
  }

  /**
   * Processes the plugins form element.
   *
   * @param array $element
   *   The form element to process.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param array $complete_form
   *   The complete form structure.
   *
   * @return array
   *   The processed element.
   */
  public static function processPluginsElement(array &$element, FormStateInterface $form_state, array &$complete_form): array {
    // Set some default values.
    if (!isset($element['#cardinality'])) {
      $element['#cardinality'] = -1;
    }
    if (!isset($element['#plugin_manager_instance'])) {
      $element['#plugin_manager_instance'] = \Drupal::service($element['#plugin_manager']);
    }

    if ((!isset($form_state->getUserInput()['_drupal_ajax']) && !$form_state->getTriggeringElement())
      || ($form_state->getTriggeringElement() && $form_state->getTriggeringElement()['#ajax']['callback'][0] != static::class)) {
      $plugin_definitions = $element['#value'] ?: $element['#default_value'];
    }
    // If we are on an ajax request coming from this object, take the data from
    // form state instead.
    else {
      $plugin_definitions = $form_state->getValue($element['#parents']);
    }

    $element['#plugins'] = [];
    foreach ($plugin_definitions as $uuid => $plugin_config) {
      if (isset($plugin_config['id']) && $element['#plugin_manager_instance']->hasDefinition($plugin_config['id'])) {
        $plugin = $element['#plugin_manager_instance']->createInstance($plugin_config['id'], $plugin_config ?? []);
        $element['#plugins'][$uuid] = $plugin;
      }
    }

    // If we have plugins, move the title to the plugin_group.
    $title = $element['#title'];
    unset($element['#title']);

    $wrapper_id = Html::getId('plugins-form-element-' . $element['#name'] . '-' . implode('_', $element['#parents']) . '-wrapper');
    return array_merge($element, [
      '#tree' => TRUE,
      '#wrapper_id' => $wrapper_id,
      '#prefix' => '<div id="' . $wrapper_id . '">',
      '#suffix' => '</div>',
      '#attached' => [
        'library' => [
          'plugin_form_element/drupal.plugins.element',
        ],
      ],
      'plugin_group' => [
        '#type' => 'vertical_tabs',
        '#title' => $title,
      ],
    ]);
  }

  /**
   * Builds the plugins form for the processed element.
   *
   * @param array $element
   *   The form element to process.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param array $complete_form
   *   The complete form structure.
   *
   * @return array
   *   The processed element.
   */
  public static function buildPluginsForm(array &$element, FormStateInterface $form_state, array &$complete_form): array {
    // If we are on an ajax request we assume that a new item has been added.
    $new_plugin_uuid = isset($form_state->getTriggeringElement()['#ajax']) ? array_key_last($element['#plugins']) : array_key_first($element['#plugins']);
    foreach ($element['#plugins'] as $uuid => $plugin) {
      $element[$uuid] = [
        '#type' => 'details',
        '#title' => $plugin->getPluginDefinition()['label'],
        '#group' => implode('][', $element['#parents']) . '][plugin_group',
        // This is currently broken in core, but we do it anyway.
        '#open' => $uuid === $new_plugin_uuid,
        '#attributes' => ['class' => ['plugins-plugin-configuration--' . $plugin->getPluginId()]],
      ];
      $element[$uuid]['configuration'] = [
        '#type' => 'plugin',
        '#plugin_manager' => $element['#plugin_manager'],
        '#default_value' => $plugin->getConfiguration(),
        '#parents' => array_merge($element['#parents'], [$uuid]),
      ];
      $element[$uuid]['id'] = [
        '#type' => 'hidden',
        '#default_value' => $plugin->getpluginId(),
      ];
      $element[$uuid]['actions'] = [
        '#type' => 'actions',
      ];
      $element[$uuid]['actions']['remove'] = [
        '#type' => 'submit',
        '#value' => t('Remove'),
        '#submit' => [[static::class, 'removePlugin']],
        // @todo add title attribute.
        '#depth' => -3,
        '#ajax' => [
          'callback' => [static::class, 'removePluginAjax'],
          'event' => 'click',
          'wrapper' => $element['#wrapper_id'],
          'uuid' => $uuid,
        ],
        '#name' => 'remove-plugin-' . $uuid,
        '#limit_validation_errors' => [],
      ];

      $form_state->addCleanValueKey(array_merge(
        $element['#parents'],
        [$uuid, 'actions']
      ));
    }

    $element['actions'] = [
      '#type' => 'actions',
      '#attributes' => [
        'class' => ['plugins-form-actions'],
      ],
    ];

    if (!static::isDisabled($element)) {
      $element['actions']['operations'] = [
        '#type' => 'plugins_operations',
        // Even though operations are run through the "links" element type, the
        // theme system will render any render array passed as a link "title".
        '#links' => static::getPluginOperationLinks($element),
        '#dropbutton_type' => 'small',
      ];
      $element['actions']['operations']['#suffix'] = t('to %type', ['%type' => $element['plugin_group']['#title']]);
    }

    // The addCleanValueKey method actually supports arrays (#3321409).
    $form_state->addCleanValueKey(array_merge($element['#parents'], ['plugin_group']));
    $form_state->addCleanValueKey(array_merge($element['#parents'], ['actions']));

    return $element;
  }

  /**
   * Retrieves plugin operation links.
   *
   * @param array $element
   *   The form element.
   *
   * @return array
   *   The operation links.
   */
  protected static function getPluginOperationLinks(array &$element): array {
    $options = array_map(
      function ($d) {
        return $d['label'];
      },
      $element['#plugin_manager_instance'] instanceof FilteredPluginManagerInterface
        ? $element['#plugin_manager_instance']->getFilteredDefinitions('plugin_config')
        : $element['#plugin_manager_instance']->getDefinitions()
    );

    foreach ($options as $key => $value) {
      $element['actions'][$key] = [
        '#type' => 'submit',
        '#value' => t('Add @plugin', ['@plugin' => $value]),
        '#submit' => [[static::class, 'addPlugin']],
        '#name' => implode('_', array_merge($element['#parents'], [$key . '_add_plugin'])),
        '#attributes' => ['class' => ['field-add-more-submit', 'button--small']],
        '#plugin_id' => $key,
        '#depth' => -2,
        '#ajax' => [
          'callback' => [static::class, 'addPluginAjax'],
          'wrapper' => $element['#wrapper_id'],
          'effect' => 'fade',
          // Since a normal throbber is added inline, this has the potential to
          // break a layout if the button is located in dropbuttons. Instead,
          // it's safer to just show the fullscreen progress element instead.
          'progress' => ['type' => 'fullscreen'],
        ],
        '#limit_validation_errors' => [$element['#parents']],
      ];
      $element['actions'][$key]['#name'] = Html::getId($element['#wrapper_id'] . '_' . $element['actions'][$key]['#name']);
      $element['actions'][$key]['#id'] = $element['actions'][$key]['#name'];
    }

    $operations = [];
    // Because we are cloning the elements into title sub element we need to
    // sort children first.
    foreach (Element::children($element['actions'], TRUE) as $child) {
      // Clone the element as an operation.
      $operations[$child] = ['title' => $element['actions'][$child]];

      // Flag the original element as printed so it doesn't render twice.
      $element['actions'][$child]['#printed'] = TRUE;
    }

    return $operations;
  }

  /**
   * {@inheritdoc}
   */
  public static function valueCallback(&$element, $input, FormStateInterface $form_state) {
    if (isset($input['plugin_group'])) {
      unset($input['plugin_group']);
    }

    return $input ?: [];
  }

  /**
   * Handles adding a new plugin to the form.
   *
   * @param array $form
   *   The form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state object.
   */
  public static function addPlugin(array &$form, FormStateInterface $form_state) {
    $form_state->cleanValues();
    $plugins = NestedArray::getValue($form, static::getParents($form_state, TRUE))['#value'] ?? [];
    $plugins[\Drupal::service('uuid')->generate()] = [
      'id' => $form_state->getTriggeringElement()['#plugin_id'],
    ];
    $form_state->setValue(static::getParents($form_state), $plugins);
    $form_state->setRebuild();
  }

  /**
   * AJAX callback for adding a plugin.
   *
   * @param array $form
   *   The form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state object.
   *
   * @return array
   *   The updated form fragment.
   */
  public static function addPluginAjax(array &$form, FormStateInterface $form_state): array {
    return NestedArray::getValue($form, static::getParents($form_state, TRUE));
  }

  /**
   * Handles removing a plugin from the form.
   *
   * @param array $form
   *   The form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state object.
   */
  public static function removePlugin(array &$form, FormStateInterface $form_state) {
    $form_state->cleanValues();
    $plugins = NestedArray::getValue($form, static::getParents($form_state, TRUE))['#value'];
    unset($plugins[$form_state->getTriggeringElement()['#ajax']['uuid']]);
    $form_state->setValue(static::getParents($form_state), $plugins);
    $form_state->setRebuild();
  }

  /**
   * AJAX callback for removing a plugin.
   *
   * @param array $form
   *   The form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state object.
   *
   * @return array
   *   The updated form fragment.
   */
  public static function removePluginAjax(array &$form, FormStateInterface $form_state): array {
    return NestedArray::getValue($form, static::getParents($form_state, TRUE));
  }

  /**
   * Determines if adding more plugins is disabled.
   *
   * @param array $element
   *   The form element.
   *
   * @return bool
   *   TRUE if adding plugins is disabled, FALSE otherwise.
   */
  protected static function isDisabled(array $element): bool {
    return $element['#cardinality'] !== -1 && count($element['#plugins']) >= $element['#cardinality'];
  }

  /**
   * Gets the parent elements for form state operations.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   * @param bool $array_parents
   *   Whether to retrieve array parents.
   *
   * @return array|null
   *   The parent elements or NULL.
   */
  protected static function getParents(FormStateInterface $form_state, bool $array_parents = FALSE): ?array {
    if ($triggering_element = $form_state->getTriggeringElement()) {
      return array_slice(
        $array_parents ? $triggering_element['#array_parents'] : $triggering_element['#parents'],
        0,
        $triggering_element['#depth']
      );
    }

    return NULL;
  }

}
