<?php

namespace Drupal\Tests\plugin_form_element\Unit;

use Drupal\Component\Plugin\PluginInspectionInterface;
use Drupal\Component\Plugin\PluginManagerInterface;
use Drupal\Tests\UnitTestCase;
use Drupal\plugin_form_element\Element\Plugin;
use Drupal\Core\Form\FormState;
use Drupal\Core\Messenger\MessengerInterface;
use PHPUnit\Framework\MockObject\MockObject;
use Symfony\Component\DependencyInjection\Container;
use Drupal\Core\StringTranslation\TranslatableMarkup;

/**
 * Unit test for the Plugin form element.
 *
 * @group plugin_form_element
 */
class PluginFormElementUnitTest extends UnitTestCase {

  /**
   * Tests error handling for missing plugin ID.
   */
  public function testErrorHandlingMissingPluginId() {
    $this->setUpMockedContainerWithMessage('The plugin form element must at least define an id');
    $this->assertPluginNotInitialized([
      '#type' => 'plugin',
      '#plugin_manager' => 'invalid.plugin.manager',
      '#default_value' => ['id' => NULL],
    ]);
  }

  /**
   * Tests error handling for missing plugin manager.
   */
  public function testErrorHandlingMissingPluginManager() {
    $this->setUpMockedContainerWithMessage('The plugin form element must define a valid plugin manager');
    $this->assertPluginNotInitialized([
      '#type' => 'plugin',
      // Simulate a missing plugin manager.
      '#plugin_manager' => NULL,
      '#default_value' => ['id' => 'some_id'],
    ]);
  }

  /**
   * Tests error handling for invalid plugin ID.
   */
  public function testErrorHandlingInvalidPluginId() {
    $this->setUpMockedContainerWithMessage('has no plugin with id');

    $plugin_manager = $this->createMock(PluginManagerInterface::class);
    $plugin_manager->method('hasDefinition')->willReturn(FALSE);
    $container = \Drupal::getContainer();
    $container->set('plugin.manager.invalid', $plugin_manager);

    $this->assertPluginNotInitialized([
      '#type' => 'plugin',
      '#plugin_manager' => 'plugin.manager.invalid',
      '#default_value' => ['id' => 'invalid_id'],
    ]);
  }

  /**
   * Tests error handling for plugin not implementing PluginFormInterface.
   */
  public function testErrorHandlingPluginNotImplementingPluginFormInterface() {
    $this->setUpMockedContainerWithMessage('is not implementing \Drupal\Core\Plugin\PluginFormInterface');

    $plugin_manager = $this->createMock(PluginManagerInterface::class);
    $plugin_manager->method('hasDefinition')->willReturn(TRUE);
    $plugin_manager->method('createInstance')->willReturn($this->createMock(PluginInspectionInterface::class));
    $container = \Drupal::getContainer();
    $container->set('plugin.manager.valid', $plugin_manager);

    $this->assertPluginNotInitialized([
      '#type' => 'plugin',
      '#plugin_manager' => 'plugin.manager.valid',
      '#default_value' => ['id' => 'valid_id'],
    ]);
  }

  /**
   * Sets up the mocked container with a messenger.
   *
   * @param \PHPUnit\Framework\MockObject\MockObject $messenger
   *   The mocked messenger.
   */
  protected function setUpMockedContainer(MockObject $messenger) {
    $container = new Container();
    $container->set('messenger', $messenger);
    \Drupal::setContainer($container);
  }

  /**
   * Mocks the messenger service and sets up the container.
   *
   * @param string $expectedMessage
   *   The expected message substring.
   *
   * @return \PHPUnit\Framework\MockObject\MockObject
   *   The mocked messenger service.
   */
  protected function setUpMockedContainerWithMessage($expectedMessage) {
    $messenger = $this->createMock(MessengerInterface::class);
    $messenger->expects($this->once())
      ->method('addWarning')
      ->with($this->callback(function ($message) use ($expectedMessage) {
        return $message instanceof TranslatableMarkup &&
          strpos($message->getUntranslatedString(), $expectedMessage) !== FALSE;
      }));

    $container = new Container();
    $container->set('messenger', $messenger);
    \Drupal::setContainer($container);

    return $messenger;
  }

  /**
   * Asserts that a plugin is not initialized during form element processing.
   *
   * This helper method processes a plugin form element and asserts that the
   * '#plugin' key is not present in the processed element. It is useful for
   * scenarios where the plugin initialization should fail or be skipped due to
   * invalid or missing configuration.
   *
   * @param array $element
   *   The plugin form element to be processed.
   */
  protected function assertPluginNotInitialized(array $element) {
    $form_state = new FormState();
    $complete_form = [];
    $processed_element = Plugin::processPluginElement($element, $form_state, $complete_form);
    $this->assertArrayNotHasKey('#plugin', $processed_element);
  }

}
